"""Aito `Data API <https://aito.ai/docs/api/#database-api>`__ Request Class"""

import re
from abc import ABC, abstractmethod
from typing import Optional, Union, Dict, List

from aito.client import responses as aito_resp
from .aito_request import AitoRequest, _FinalRequest, _PatternEndpoint, _GetRequest, _PostRequest
from aito.schema import AitoSchema
from ..responses.query_api_response import ModifyResponse


class DataAPIRequest(AitoRequest, ABC):
    """Request to manipulate the schema"""
    endpoint_prefix = f'{AitoRequest._api_version_endpoint_prefix}/data'
    # Partial list of available endpoints.
    _data_api_jobs_methods = ['optimize', 'batch']

    @classmethod
    @abstractmethod
    def _check_method(cls, method: str) -> bool:
        return method in cls._request_methods

    @classmethod
    @abstractmethod
    def _check_endpoint(cls, endpoint: str) -> bool:
        return endpoint.startswith(cls.endpoint_prefix)

    @classmethod
    @abstractmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        for sub_cls in cls.__subclasses__():
            if sub_cls._check_method(method) and sub_cls._check_endpoint(endpoint):
                return sub_cls.make_request(method=method, endpoint=endpoint, query=query)
        raise ValueError(f"invalid {cls.__name__} with '{method}({endpoint})'")

    @classmethod
    @abstractmethod
    def path_matches(cls, path_to_match):
        return False

class UploadEntriesRequest(_PostRequest, _PatternEndpoint, DataAPIRequest):
    """Request to `Insert entries to a table <https://aito.ai/docs/api/#post-api-v1-data-table>`__"""
    _path_suffix = 'batch'
    response_cls = aito_resp.BaseResponse

    @classmethod
    def _endpoint_pattern(cls):
        return re.compile(f'^{cls.endpoint_prefix}/({AitoSchema.table_name_pattern})/{cls._path_suffix}$')

    def __init__(self, table_name: str, entries: List[Dict]):
        """

        :param table_name: the name of the table to be uploaded
        :type table_name: str
        :param entries: a list of the table entries
        :type entries: List[Dict]
        """
        endpoint = f'{self.endpoint_prefix}/{table_name}/batch'
        super().__init__(method=self.method, endpoint=endpoint, query=entries)

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        groups = cls._endpoint_to_captured_groups(endpoint=endpoint)
        return cls(table_name=groups(1), entries=query)

    @classmethod
    def path_matches(cls, path_to_match):
        return path_to_match.endswith(f'/{cls._path_suffix}')


class OptimizeTableRequest(_PostRequest, _FinalRequest, DataAPIRequest):
    """Request to `Optimize a table <https://aito.ai/docs/api/#post-api-v1-data-table-optimize>`__"""
    _path_suffix = 'optimize'

    def __init__(self, table_name: str, session_id: str):
        """

        :param table_name: the name of the table to be optimized
        :type table_name: str
        """
        endpoint = f'{self.endpoint_prefix}/{table_name}/optimize'
        super().__init__(method=self.method, endpoint=endpoint)

    @classmethod
    def _endpoint_pattern(cls):
        return re.compile(f'^{cls.endpoint_prefix}/{AitoSchema.table_name_pattern}/{cls._path_suffix}$')

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        groups = cls._endpoint_to_captured_groups(endpoint=endpoint)
        return cls(table_name=groups(1), session_id=groups(2))

    @classmethod
    def path_matches(cls, path_to_match):
        return path_to_match.endswith(f'/{cls._path_suffix}')

    response_cls = aito_resp.BaseResponse

class DeleteEntriesRequest(_PostRequest, _FinalRequest, DataAPIRequest):
    """Request to `Delete entries of a table <https://aito.ai/docs/api/#post-api-v1-data-delete>`__"""
    _path_suffix = '_delete'

    @classmethod
    def _endpoint_pattern(cls):
        return re.compile(f'^{cls.endpoint_prefix}/{cls._path_suffix}$')

    @classmethod
    def path_matches(cls, path_to_match):
        return path_to_match.endswith(f'/{cls._path_suffix}')


    endpoint = f'{DataAPIRequest.endpoint_prefix}/{_path_suffix}'
    response_cls = aito_resp.BaseResponse


class InitiateFileUploadRequest(_PostRequest, _PatternEndpoint, DataAPIRequest):
    """Request to `Initiate File Upload <https://aito.ai/docs/api/#post-api-v1-data-table-file>`__"""
    _path_suffix = 'file'
    response_cls = aito_resp.BaseResponse

    @classmethod
    def _endpoint_pattern(cls):
        return re.compile(f'^{cls.endpoint_prefix}/({AitoSchema.table_name_pattern})/{cls._path_suffix}$')

    def __init__(self, table_name: str):
        """

        :param table_name: the name of the table to be uploaded
        :type table_name: str
        """
        endpoint = f'{self.endpoint_prefix}/{table_name}/{self._path_suffix}'
        super().__init__(method=self.method, endpoint=endpoint)

    @classmethod
    def path_matches(cls, path_to_match):
        return path_to_match.endswith(f'/{cls._path_suffix}')

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        groups = cls._endpoint_to_captured_groups(endpoint=endpoint)
        return cls(table_name=groups(1))


class TriggerFileProcessingRequest(_PostRequest, _PatternEndpoint, DataAPIRequest):
    """Request to `Initiate File Upload <https://aito.ai/docs/api/#post-api-v1-data-table-file>`__"""
    response_cls = aito_resp.BaseResponse

    @classmethod
    def _endpoint_pattern(cls):
        return re.compile(f'^{cls.endpoint_prefix}/({AitoSchema.table_name_pattern})/file/({AitoSchema.uuid_pattern})$')

    def __init__(self, table_name: str, session_id: str):
        """

        :param table_name: the name of the table to be uploaded
        :type table_name: str
        :param session_id: The uuid of the file upload session from initiating file upload
        :type session_id: str
        """
        endpoint = f'{self.endpoint_prefix}/{table_name}/file/{session_id}'
        super().__init__(method=self.method, endpoint=endpoint)

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        groups = cls._endpoint_to_captured_groups(endpoint=endpoint)
        return cls(table_name=groups(1), session_id=groups(2))

    @classmethod
    def path_matches(cls, path_to_match):
        matcher = re.compile(f'file/{AitoSchema.uuid_pattern}$')
        return bool(matcher.match(path_to_match))



class GetFileProcessingRequest(_GetRequest, TriggerFileProcessingRequest, DataAPIRequest):
    """Request to `Initiate File Upload <https://aito.ai/docs/api/#post-api-v1-data-table-file>`__"""


class ModifyRequest(_PostRequest, _FinalRequest, DataAPIRequest):
    """Request to `Modify data atomically <https://aito.ai/docs/api/#post-api-v1-data-modify>`__

    Allows individual modifications or a sequence of modifications in one atomic operation.
    """
    _path_suffix = '_modify'
    endpoint = f'{DataAPIRequest.endpoint_prefix}/{_path_suffix}'
    response_cls = ModifyResponse

    def __init__(self, query: Dict):
        """
        :param query: The modify query with operations
        :type query: Dict
        """
        super().__init__(query=query)

    @classmethod
    def _endpoint_pattern(cls):
        return re.compile(f'^{cls.endpoint}$')

    @classmethod
    def path_matches(cls, path_to_match):
        return path_to_match.endswith(f'/{cls._path_suffix}')



"""Aito `Query API <https://aito.ai/docs/api/#query-api>`__ Request Class"""

import re
from abc import ABC
from typing import Dict, Optional, Union, List

from .aito_request import AitoRequest, _PatternEndpoint, _PostRequest, _FinalRequest
from ..responses import SearchResponse, PredictResponse, RecommendResponse, EvaluateResponse, SimilarityResponse, \
    MatchResponse, RelateResponse, HitsResponse, EstimateResponse, AggregateResponse, BatchResponse


class QueryAPIRequest(_PostRequest, _PatternEndpoint, AitoRequest, ABC):
    """Request to a `Query API <https://aito.ai/docs/api/#query-api>`__
    """
    #: the Query API path
    path: str = None  # get around of not having abstract class attribute
    _query_api_paths = ['_search', '_predict', '_recommend', '_evaluate', '_similarity', '_match', '_relate', '_query', '_estimate', '_aggregate']

    def __init__(self, query: Dict):
        """

        :param query: an Aito query if applicable, optional
        :type query: Dict
        """
        if self.path is None:
            raise NotImplementedError(f'The API path must be implemented')
        endpoint = self._endpoint_from_path(self.path)
        super().__init__(method=self.method, endpoint=endpoint, query=query)

    @classmethod
    def _endpoint_pattern(cls):
        return re.compile(f"^{cls._api_version_endpoint_prefix}/({'|'.join(cls._query_api_paths)})$")

    @classmethod
    def make_request(cls, method: str, endpoint: str, query: Optional[Union[Dict, List]]) -> 'AitoRequest':
        for sub_cls in cls.__subclasses__():
            if method == sub_cls.method and endpoint == QueryAPIRequest._endpoint_from_path(sub_cls.path):
                return sub_cls(query=query)
        raise ValueError(f"invalid {cls.__name__} with '{method}({endpoint})'")

    @classmethod
    def _endpoint_from_path(cls, path: str):
        """return the query api endpoint from the Query API path"""
        if path not in cls._query_api_paths:
            raise ValueError(f"path must be one of {'|'.join(cls._query_api_paths)}")
        return f'{cls._api_version_endpoint_prefix}/{path}'


class SearchRequest(QueryAPIRequest):
    """Request to the `Search API <https://aito.ai/docs/api/#post-api-v1-search>`__"""

    #: the Query API path
    path: str = '_search'
    response_cls = SearchResponse


class PredictRequest(QueryAPIRequest):
    """Request to the `Predict API <https://aito.ai/docs/api/#post-api-v1-predict>`__"""
    #: the Query API path
    path: str = '_predict'
    #: the class of the response for this request class
    response_cls = PredictResponse


class RecommendRequest(QueryAPIRequest):
    """Request to the `Recommend API <https://aito.ai/docs/api/#post-api-v1-recommend>`__"""
    #: the Query API path
    path: str = '_recommend'
    response_cls = RecommendResponse


class EvaluateRequest(QueryAPIRequest):
    """Request to the `Evaluate API <https://aito.ai/docs/api/#post-api-v1-evaluate>`__"""
    #: the Query API path
    path: str = '_evaluate'
    response_cls = EvaluateResponse


class SimilarityRequest(QueryAPIRequest):
    """Request to the `Similarity API <https://aito.ai/docs/api/#post-api-v1-similarity>`__"""
    #: the Query API path
    path: str = '_similarity'
    response_cls = SimilarityResponse


class MatchRequest(QueryAPIRequest):
    """Request to the `Match query <https://aito.ai/docs/api/#post-api-v1-match>`__"""
    #: the Query API path
    path: str = '_match'
    response_cls = MatchResponse


class RelateRequest(QueryAPIRequest):
    """Request to the `Relate API <https://aito.ai/docs/api/#post-api-v1-relate>`__"""
    #: the Query API path
    path: str = '_relate'
    response_cls = RelateResponse


class GenericQueryRequest(QueryAPIRequest):
    """Request to the `Generic Query API <https://aito.ai/docs/api/#post-api-v1-query>`__"""
    path: str = '_query'
    response_cls = HitsResponse


class EstimateRequest(QueryAPIRequest):
    """Request to the `Estimate API <https://aito.ai/docs/api/#post-api-v1-estimate>`__"""
    #: the Query API path
    path: str = '_estimate'
    #: the class of the response for this request class
    response_cls = EstimateResponse


class AggregateRequest(QueryAPIRequest):
    """Request to the `Aggregate API <https://aito.ai/docs/api/#post-api-v1-aggregate>`__"""
    #: the Query API path
    path: str = '_aggregate'
    #: the class of the response for this request class
    response_cls = AggregateResponse


class BatchRequest(_PostRequest, _FinalRequest, AitoRequest):
    """Request to the `Batch API <https://aito.ai/docs/api/#post-api-v1-batch>`__

    Allows executing multiple queries (predict, similarity, search, etc.) in a single request.

    Example::

        from aito.client.requests import BatchRequest

        # Execute multiple queries in one request
        request = BatchRequest([
            {"from": "products", "where": {"name": "rye bread"}, "predict": "category"},
            {"from": "products", "similarity": {"name": "rye bread"}}
        ])
    """
    endpoint = f'{AitoRequest._api_version_endpoint_prefix}/_batch'
    response_cls = BatchResponse

    def __init__(self, queries: List[Dict]):
        """
        :param queries: A list of query dictionaries to execute in batch
        :type queries: List[Dict]
        """
        super().__init__(query=queries)


"""Helper classes for Aito Modify API operations.

These classes provide a fluent interface for building modify queries::

    from aito.client.requests import Insert, Update, Delete
    import aito.api as api

    # Insert a single entry
    api.modify(client, Insert("products", {"id": "1", "name": "Apple"}))

    # Insert multiple entries
    entries = [{"id": "1", "name": "Apple"}, {"id": "2", "name": "Banana"}]
    api.modify(client, Insert("products", entries))

    # Update entries matching a condition
    api.modify(client, Update("products").where({"id": "1"}).set({"name": "New Name"}))

    # Update with upsert (insert if not exists)
    api.modify(client, Update("products").where({"id": "1"}).set({"name": "Name"}).upsert())

    # Delete entries matching a condition
    api.modify(client, Delete("products", {"id": "1"}))

    # Multiple operations atomically
    ops = [Insert("products", {"id": "1", "name": "Apple"}),
           Update("products").where({"id": "2"}).set({"name": "Banana"}),
           Delete("products", {"id": "3"})]
    api.modify(client, ops)
"""

from abc import ABC, abstractmethod
from typing import Dict, List, Union


class ModifyOperation(ABC):
    """Base class for modify operations."""

    @abstractmethod
    def to_query(self) -> Dict:
        """Convert the operation to a query dictionary."""
        pass


class Insert(ModifyOperation):
    """Insert operation for the Modify API.

    Example::

        # Insert single entry
        Insert("products", {"id": "1", "name": "Apple"})

        # Insert multiple entries
        Insert("products", [{"id": "1"}, {"id": "2"}])
    """

    def __init__(self, table: str, entries: Union[Dict, List[Dict]]):
        """
        :param table: The name of the table to insert into
        :type table: str
        :param entries: A single entry dict or a list of entry dicts to insert
        :type entries: Union[Dict, List[Dict]]
        """
        self.table = table
        self.entries = entries

    def to_query(self) -> Dict:
        return {
            "into": self.table,
            "insert": self.entries
        }


class Update(ModifyOperation):
    """Update operation for the Modify API.

    Uses a fluent interface::

        Update("products").where({"id": "1"}).set({"name": "New Name"})

        # With upsert (insert if not exists)
        Update("products").where({"id": "1"}).set({"name": "Name"}).upsert()
    """

    def __init__(self, table: str):
        """
        :param table: The name of the table to update
        :type table: str
        """
        self._table = table
        self._where: Dict = None
        self._set: Dict = None
        self._upsert: bool = False

    def where(self, condition: Dict) -> 'Update':
        """Set the condition for which entries to update.

        :param condition: The condition to match entries
        :type condition: Dict
        :return: self for chaining
        :rtype: Update
        """
        self._where = condition
        return self

    def set(self, fields: Dict) -> 'Update':
        """Set the fields to update.

        :param fields: The fields and values to set
        :type fields: Dict
        :return: self for chaining
        :rtype: Update
        """
        self._set = fields
        return self

    def upsert(self, enable: bool = True) -> 'Update':
        """Enable upsert mode (insert if not exists).

        :param enable: Whether to enable upsert, defaults to True
        :type enable: bool
        :return: self for chaining
        :rtype: Update
        """
        self._upsert = enable
        return self

    def to_query(self) -> Dict:
        if self._where is None:
            raise ValueError("Update operation requires a where() condition")
        if self._set is None:
            raise ValueError("Update operation requires set() fields")

        query = {
            "update": self._table,
            "where": self._where,
            "set": self._set
        }
        if self._upsert:
            query["upsert"] = True
        return query


class Delete(ModifyOperation):
    """Delete operation for the Modify API.

    Example::

        Delete("products", {"id": "1"})
    """

    def __init__(self, table: str, condition: Dict):
        """
        :param table: The name of the table to delete from
        :type table: str
        :param condition: The condition to match entries to delete
        :type condition: Dict
        """
        self.table = table
        self.condition = condition

    def to_query(self) -> Dict:
        return {
            "from": self.table,
            "delete": self.condition
        }
